/****************************************************************************
*
*	Example Code for MNM1221
*
*	Copyright (c) 2006-2023 Panasonic Industry Co., Ltd.
*
*----------------------------------------------------------------------------
*
*	File Name:		mnm1221_s.c
*	Description:	Example code to access MNM1221 as SLAVE mode
*   Notes:
*	- This code is just for reference, so that Panasonic does NOT assume 
*	  responsibility or liability for actual use based upon this code.
*	- This file is NOT for warranty of correct behavior, 
*	  and might include bugs, so that you have to be careful.
*	- The codes are very basic only, therefore you will need additional code 
*	  such as the process at the occurrence of errors.
*	- It is assumed that TAB of your PC is set to 4space width.
*
*----------------------------------------------------------------------------
*	Revision:	Date(y/m/d):	Author:			Comments:
*	0.1			2006/02/02		Y. Fujisaki		Initial release
*	0.2			2023/03/31		Y. Fujisaki		Changed company name
*
*****************************************************************************/

#include	"mnm1221_s.h"	/* Modify path according to your file structure */


/****************************************************************************/
/* Definitions and Declarations												*/
/****************************************************************************/
/*
* Some of the followings accessed in also other files should be moved to 
* the header file you provide separately.
*/

/*--------------------------------------------------------------------------*/
/* Definition of constants													*/
/*--------------------------------------------------------------------------*/

#define	N_MY_BLOCK		1		/* the number of my data blocks */
								/* setting range: 1 to 32, normally 1 */
#if N_MY_BLOCK >= 32
	#define	MSK_DATA_CRCE	0xffffffff
#else
	#define	MSK_DATA_CRCE	(((unsigned long)1 << N_MY_BLOCK) - 1)
#endif

#define	N_MY_DATA		(SIZE_OF_BLOCK * N_MY_BLOCK)


/*--------------------------------------------------------------------------*/
/* Declaration and definition of variables used in also other files			*/
/*--------------------------------------------------------------------------*/
/*
* If your compiler does not allow that a file includes both declaration and 
* definition of variables, the declaration should be moved to the other file.
*/

/* declaration */

#ifdef SLAVE_16BIT_ACCESS	/* 16bits bus */
	extern unsigned short my_rx_buf[]; 			/* RX data buffer */
	extern unsigned short my_tx_buf[]; 			/* TX data buffer */
#else						/* 8bits bus */
	extern unsigned char my_rx_buf[]; 			/* RX data buffer */
	extern unsigned char my_tx_buf[]; 			/* TX data buffer */
#endif

extern unsigned short crc_err_count;			/* CRC error counter */

/* definition */

#ifdef SLAVE_16BIT_ACCESS	/* 16bits bus */
	unsigned short my_rx_buf[N_MY_DATA]; 		/* RX data buffer */
	unsigned short my_tx_buf[N_MY_DATA]; 		/* TX data buffer */
#else						/* 8bits bus */
	unsigned char my_rx_buf[N_MY_DATA]; 		/* RX data buffer */
	unsigned char my_tx_buf[N_MY_DATA]; 		/* TX data buffer */
#endif

unsigned short crc_err_count;					/* CRC error counter */


/*--------------------------------------------------------------------------*/
/* Definition of local variables											*/
/*--------------------------------------------------------------------------*/

#ifdef SLAVE_16BIT_ACCESS	/* 16bits bus */
	static volatile unsigned short *p_my_rx_bgn;	/* pointer to access RX data */
#else						/* 8bits bus */
	static volatile unsigned char *p_my_rx_bgn;		/* pointer to access RX data */
#endif

static short my_mac_id;			/* MAC-ID (Node Address) */
static short done_1st_cycle;	/* flag to detect the first cycle in RUNNING state */


/*--------------------------------------------------------------------------*/
/* Declaration (prototype) of functions										*/
/*--------------------------------------------------------------------------*/

/* global (called in other files)  */
extern void init_mnm1221_s(short mac_id);
extern short get_state_mnm1221_s(void);
extern short int_sync_mnm1221_s(void);
extern void timeout_mnm1221_s(short mac_id);

/* local */
static void clr_com_buf(void);
static void clr_mnm1221_tx_mem(void);
static short chk_crc_err(unsigned long mask);

/* local and general */
static void mem_copy_s(volatile unsigned short *ptr1, volatile unsigned short *ptr2, short n);
static void mem_copy_c(volatile unsigned char *ptr1, volatile unsigned char *ptr2, short n);

/* for test only */
static short set_resp_example(short mac_id, short crc_err);


/*--------------------------------------------------------------------------*/
/* Definition of local macros												*/
/*--------------------------------------------------------------------------*/

#ifdef SLAVE_16BIT_ACCESS
	#define	get_cmd_mnm1221()	mem_copy_s(&my_rx_buf[0], p_my_rx_bgn, N_MY_DATA)
	#define	set_rsp_mnm1221()	mem_copy_s(P_MNM1221_TX_MEM_BGN, &my_tx_buf[0], N_MY_DATA)
#else
	#define	get_cmd_mnm1221()	mem_copy_c(&my_rx_buf[0], p_my_rx_bgn, N_MY_DATA)
	#define	set_rsp_mnm1221()	mem_copy_c(P_MNM1221_TX_MEM_BGN, &my_tx_buf[0], N_MY_DATA)
#endif


/****************************************************************************/
/* Global Functions															*/
/****************************************************************************/

/*----------------------------------------------------------------------------
*
*	Function Name:	init_mnm1221_s
*
*	Description:	initializing MNM1221 registers and variables
*
*	Argument(s):	MAC-ID (0 to 31)
*
*	Return Value:	none
*
*	Comment:		Call after reset-release before main loop.
*					It is assumed that the timeout detection is done with firmware 
*					you provide, so the function of MNM1221 is disabled.
*
*---------------------------------------------------------------------------*/
void init_mnm1221_s(short mac_id)
{
	/* variables */
	clr_com_buf();					/* clear communication buffer */
	done_1st_cycle = 0;
	crc_err_count = 0;
	my_mac_id = mac_id & 0x001f;

	/* MNM1221 registers */
	MNM1221_S_MACID = my_mac_id;	/* MAC-ID */
	MNM1221_S_BLK_N = N_MY_BLOCK;	/* the number of my data blocks */
	MNM1221_S_TIMEOUT = 0;			/* disable timeout detection */
	MNM1221_S_INTRX_MSK = 1;		/* disable XINTRX output at timeout */
	MNM1221_S_SYNC_DELAY = 0;		/* non delay */
	MNM1221_S_SYNC_MSK = 0;			/* enable XSYNC output */
	MNM1221_S_RXMEM_HOLD = 0;		/* enable to switch RX bank */

	clr_mnm1221_tx_mem();			/* clear TX memory */

	MNM1221_S_INIT_DONE = 1;		/* enable communication */
}


/*----------------------------------------------------------------------------
*
*	Function Name:	get_state_mnm1221_s
*
*	Description:	monitoring the state of MNM1221
*
*	Argument(s):	none
*
*	Return Value:	0x0001 ... RUNNING
*					0x0002 ... CONFIG-B
*					0x0004 ... CONFIG-A
*					0x0008 ... INITIAL
*
*	Comment:		This function should be placed in the main loop or 
*					a timer interrupt which has a lower or same priority than XSYNC.
*					Using this return and checking errors, control "COM" LED.
*
*---------------------------------------------------------------------------*/
short get_state_mnm1221_s(void)
{
#ifdef SLAVE_16BIT_ACCESS
	return (MNM1221_S_STATE & 0x000f);
#else
	return (MNM1221_S_STATE_L & 0x000f);
#endif
}


/*----------------------------------------------------------------------------
*
*	Function Name:	int_sync_mnm1221_s
*
*	Description:	interrupt routine by XSYNC of MNM1221
*
*	Argument(s):	none
*
*	Return Value:	0: OK, -1: CRC error
*
*	Comment:		Call in the interrupt function by XSYNC.
*
*---------------------------------------------------------------------------*/
short int_sync_mnm1221_s(void)
{
	static unsigned long my_dcrc_mask;		/* mask to extract data-CRC flag */
	short order;							/* order of my block */
	short crc_err;							/* flag of CRC error */

	/* only at the first cycle in RUNNING state, set read-out pointer */
	if (!done_1st_cycle) {
		done_1st_cycle = 1;

		order = MNM1221_S_BLK_ORDER & 0x001f;
		p_my_rx_bgn = P_MNM1221_RX_MEM_BGN + (SIZE_OF_BLOCK * order);
		my_dcrc_mask = MSK_DATA_CRCE << order;
	}

	/* CRC error checking */
	crc_err = chk_crc_err(my_dcrc_mask);

	/* get command from MNM1221 to my_rx_buf[] */
	if (!crc_err) {
		get_cmd_mnm1221();
	}

	/*--- Here, put your application using my_rx_buf[] and my_tx_buf[] -----*/

	/* This function is just for example */
	set_resp_example(my_mac_id, crc_err);

	/*------------------------------------------ end of your application ---*/

	/* set response from my_tx_buf[] to MNM1221 */
	set_rsp_mnm1221();
	MNM1221_S_TXMEM_SW = 1;					/* switching TX bank */

	return crc_err;
}


/*----------------------------------------------------------------------------
*
*	Function Name:	timeout_mnm1221_s
*
*	Description:	operation at timeout
*
*	Argument(s):	MAC-ID (0 to 31)
*
*	Return Value:	none
*
*	Comment:		If the timeout is detected in the function you provide, 
*					this function should be called.
*
*					Because a user might change the order of the cable connection 
*					after timeout, it is necessary to reset MNM1221 
*					in order to force Master to re-start from RING-CONFIG state.
*
*---------------------------------------------------------------------------*/
void timeout_mnm1221_s(short mac_id)
{
	MNM1221_S_RESET = 1;		/* reset MNM1221 */
								/* Another way is to control an output port */
								/* connected with XRST input of MNM1221. */
	init_mnm1221_s(mac_id);

	/*------------------------------------------------------------------------------*/
	/* Here, put the operation to blink the "COM" LED (ON:0.5s, OFF:0.5s) with red. */
	/*------------------------------------------------------------------------------*/
}


/****************************************************************************/
/* Local Functions															*/
/****************************************************************************/

/*----------------------------------------------------------------------------
*
*	Function Name:	clr_com_buf
*
*	Description:	clearing communication buffer
*
*	Argument(s):	none
*
*	Return Value:	none
*
*	Comment:		
*
*---------------------------------------------------------------------------*/
static void clr_com_buf(void)
{
	short i;

	for (i = 0; i < N_MY_DATA; i++) {
		my_rx_buf[i] = 0;
		my_tx_buf[i] = 0;
	}
}


/*----------------------------------------------------------------------------
*
*	Function Name:	clr_mnm1221_tx_mem
*
*	Description:	clearing TX memory in MNM1221
*
*	Argument(s):	none
*
*	Return Value:	none
*
*	Comment:		
*
*---------------------------------------------------------------------------*/
static void clr_mnm1221_tx_mem(void)
{
	short i;

#ifdef SLAVE_16BIT_ACCESS
	volatile unsigned short *p;
#else
	volatile unsigned char *p;
#endif

/* one bank */
	p = P_MNM1221_TX_MEM_BGN;
	for (i = 0; i < N_MY_DATA; i++) {
		*p++ = 0;
	}
	MNM1221_S_TXMEM_SW = 1;				/* switching TX bank */

/* the other bank */
	p = P_MNM1221_TX_MEM_BGN;
	for (i = 0; i < N_MY_DATA; i++) {
		*p++ = 0;
	}
	MNM1221_S_TXMEM_SW = 1;				/* switching TX bank */
}


/*----------------------------------------------------------------------------
*
*	Function Name:	chk_crc_err
*
*	Description:	checking CRC error
*
*	Argument(s):	mask code for detecting
*
*	Return Value:	0: OK, -1: error
*
*	Comment:		
*
*---------------------------------------------------------------------------*/
static short chk_crc_err(unsigned long mask)
{
	unsigned long data_crc_err;

#ifdef SLAVE_16BIT_ACCESS
	data_crc_err = MNM1221_S_DCRC_ERR_H;
	data_crc_err <<= 16;
	data_crc_err |= MNM1221_S_DCRC_ERR_L;
#else
	data_crc_err = MNM1221_S_DCRC_ERRHH;
	data_crc_err <<= 8;
	data_crc_err |= MNM1221_S_DCRC_ERRHL;
	data_crc_err <<= 8;
	data_crc_err |= MNM1221_S_DCRC_ERRLH;
	data_crc_err <<= 8;
	data_crc_err |= MNM1221_S_DCRC_ERRLL;
#endif

	if (data_crc_err & mask) {	/* error */
		crc_err_count++;
		return -1;
	} else {					/* ok */
		return 0;
	}
}


/****************************************************************************/
/* General Functions														*/
/****************************************************************************/

/*----------------------------------------------------------------------------
*
*	Function Name:	mem_copy_s
*
*	Description:	copy from memory 2 to memory 1
*
*	Argument(s):	ptr1 ... pointer of memory 1
*					ptr2 ... pointer of memory 2
*					n    ... number of copying
*
*	Return Value:	none
*
*	Comment:		16bits access
*
*---------------------------------------------------------------------------*/
static void mem_copy_s(volatile unsigned short *ptr1, volatile unsigned short *ptr2, short n)
{
	for (; n > 0; n--) {
		*ptr1++ = *ptr2++;
	}
}


/*----------------------------------------------------------------------------
*
*	Function Name:	mem_copy_c
*
*	Description:	copy from memory 2 to memory 1
*
*	Argument(s):	ptr1 ... pointer of memory 1
*					ptr2 ... pointer of memory 2
*					n    ... number of copying
*
*	Return Value:	none
*
*	Comment:		8bits access
*
*---------------------------------------------------------------------------*/
static void mem_copy_c(volatile unsigned char *ptr1, volatile unsigned char *ptr2, short n)
{
	for (; n > 0; n--) {
		*ptr1++ = *ptr2++;
	}
}


/****************************************************************************/
/* Test Functions															*/
/****************************************************************************/

/*----------------------------------------------------------------------------
*
*	Function Name:	set_resp_example
*
*	Description:	an example for setting response data as echo back
*
*	Argument(s):	mac_id  ... MAC-ID (0 to 31)
*					crc_err ... OK: 0, error: not 0
*
*	Return Value:	0: OK, -1: CRC error, -2: MAC-ID error
*
*	Comment:		This function is for test only.
*					The byte 0 and 1 follow the servo protocol.
*					The byte 2 or more are echo backed regardless of any errors.
*
*---------------------------------------------------------------------------*/
#define MSK_MACID		0x001f		/* mask for MAC-ID */
#define	MSK_CMD_CD		0x7f00		/* mask for Command Code */
#define	B_CMD_RSP		0x0080		/* CMD/RSP bit */
#define	B_CMD_ERR		0x8000		/* CMD error bit */
#define	CRC_ERR_RSP		(B_CMD_ERR | MSK_CMD_CD)	/* CRC error response code */

static short set_resp_example(short mac_id, short crc_err)
{
	unsigned short byte1_0;
	short err;

#ifdef SLAVE_16BIT_ACCESS
	byte1_0 = my_rx_buf[0];
#else
	byte1_0 = my_rx_buf[1];
	byte1_0 <<= 8;
	byte1_0 |= my_rx_buf[0];
#endif

	byte1_0 &= (~MSK_MACID);
	byte1_0 |= mac_id;
	byte1_0 |= B_CMD_RSP;

	if (crc_err) {
	/* CRC error */
		err = -1;
		byte1_0 |= CRC_ERR_RSP;

	} else if ((my_rx_buf[0] & MSK_MACID) != mac_id) {
	/* MAC-ID error */
		err = -2;
		byte1_0 |= B_CMD_ERR;

	} else {
		err = 0;
	}

#ifdef SLAVE_16BIT_ACCESS
	my_tx_buf[0] = byte1_0;
	mem_copy_s(&my_tx_buf[1], &my_rx_buf[1], (N_MY_DATA - 1));	/* echo back */
#else
	my_tx_buf[0] = byte1_0;
	my_tx_buf[1] = byte1_0 >> 8;
	mem_copy_c(&my_tx_buf[2], &my_rx_buf[2], (N_MY_DATA - 2));	/* echo back */
#endif

	return err;
}


/***************************************************** End of mnm1221_s.c ***/

